import os
from .ass_style import AssStyle
from .usage_data import UsageData
from ass import parse_file, parse_string, Dialogue, Document
from ass_tag_analyzer import (
    AssInvalidTagBold,
    AssInvalidTagFontName,
    AssInvalidTagItalic,
    AssInvalidTagResetStyle,
    AssInvalidTagWrapStyle,
    AssItem,
    AssTagBold,
    AssTagFontName,
    AssTagItalic,
    AssTagResetStyle,
    AssTagWrapStyle,
    AssText,
    AssValidTagAnimation,
    AssValidTagBold,
    AssValidTagFontName,
    AssValidTagItalic,
    AssValidTagResetStyle,
    AssValidTagWrapStyle,
    parse_line,
    WrapStyle,
)
from typing import Dict, List


class AssDocument:

    subtitle: Document

    def __init__(self, subtitle: Document):
        self.subtitle = subtitle

    @classmethod
    def from_string(cls, file_text: str):
        return cls(parse_string(file_text))

    @classmethod
    def from_file(cls, filename: str, encoding: str = "utf_8_sig"):
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"The file {filename} is not reachable")

        with open(filename, encoding=encoding) as file:
            subtitle = parse_file(file)

        return cls(subtitle)

    def _set_used_styles(
        self,
        used_styles: Dict[AssStyle, UsageData],
        tags: List[AssItem],
        line_index: int,
        sub_styles: Dict[str, AssStyle],
        original_line_style: AssStyle,
        line_style: AssStyle,
        current_style: AssStyle,
        current_wrap_style: WrapStyle,
    ) -> None:
        """
        Parameters:
            used_styles (Dict[AssStyle, UsageData]): This variable will be modified
            tags (List[AssItem]): List of all tags
            line_index (int): Position of the line in the subtitle
            sub_styles (Dict[str, AssStyle]): Dict of the [V4+ Styles] sections
            original_line_style (AssStyle): Style of the line
            line_style (AssStyle): Style of the line. In general, it will be equal to original_line_style except it there is an \rXXX
            current_style (AssStyle): Real style of the text. It exist since \fn, \b, \i can override the line_style.
            current_wrap_style (WrapStyle): Since \q can override the subtitle WrapStyle, we need it.
        """

        for tag in tags:
            if isinstance(tag, AssTagResetStyle):
                if isinstance(tag, AssValidTagResetStyle):
                    style = sub_styles.get(tag.style, original_line_style)

                    # Copy the style
                    line_style = AssStyle(style.fontname, style.weight, style.italic)
                    current_style = AssStyle(style.fontname, style.weight, style.italic)
                elif isinstance(tag, AssInvalidTagResetStyle):
                    # Copy the original_line_style
                    line_style = AssStyle(
                        original_line_style.fontname,
                        original_line_style.weight,
                        original_line_style.italic,
                    )
                    current_style = AssStyle(
                        original_line_style.fontname,
                        original_line_style.weight,
                        original_line_style.italic,
                    )

            elif isinstance(tag, AssTagBold):
                if isinstance(tag, AssValidTagBold):
                    current_style.weight = tag.weight
                elif isinstance(tag, AssInvalidTagBold):
                    current_style.weight = line_style.weight

            elif isinstance(tag, AssTagItalic):
                if isinstance(tag, AssValidTagItalic):
                    current_style.italic = tag.enabled
                elif isinstance(tag, AssInvalidTagItalic):
                    current_style.italic = line_style.italic

            elif isinstance(tag, AssTagFontName):
                if isinstance(tag, AssValidTagFontName):
                    current_style.fontname = tag.name

                elif isinstance(tag, AssInvalidTagFontName):
                    current_style.fontname = line_style.fontname

            elif isinstance(tag, AssTagWrapStyle):
                if isinstance(tag, AssValidTagWrapStyle):
                    current_wrap_style = tag.style
                elif isinstance(tag, AssInvalidTagWrapStyle):
                    current_wrap_style = WrapStyle(self.subtitle.wrap_style)

            elif isinstance(tag, AssValidTagAnimation):
                self._set_used_styles(
                    used_styles,
                    tag.tags,
                    line_index,
                    sub_styles,
                    original_line_style,
                    line_style,
                    current_style,
                    current_wrap_style,
                )

            elif isinstance(tag, AssText):
                # Inspired by
                #     - https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_parse.c#L1039-L1075
                #     - Aegisub FontCollector ignore \n: https://github.com/arch1t3cht/Aegisub/blob/fad362ec2e2975d8e37893c6dfb3a39452e71d23/src/font_file_lister.cpp#L118-L120
                text = tag.text.replace("\t", " ")
                if current_wrap_style == WrapStyle.NO_WORD:
                    text = text.replace("\\n", "")
                else:
                    text = text.replace("\\n", " ")
                text = text.replace("\\N", "")
                # Libass use latin space to render NBSP: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_font.c#L573-L574
                text = text.replace("\\h", " ")
                text = text.replace("\u00A0", " ")

                # Update or create the usage_data
                usage_data = used_styles.get(current_style, None)
                if usage_data is None:
                    usage_data = UsageData(set(text), set([line_index]))
                    used_styles[current_style] = usage_data
                else:
                    usage_data.characters_used.update(set(text))
                    usage_data.lines.add(line_index)

                # We need to make an copy of the style since current_style can be modified
                current_style = AssStyle(
                    current_style.fontname, current_style.weight, current_style.italic
                )

    def get_used_style(self) -> Dict[AssStyle, UsageData]:
        """
        Returns:
            An dictionnary which contain all the used AssStyle and it's UsageData.
        """
        used_styles: Dict[AssStyle, UsageData] = {}

        # VSFilter trim:
        #   - *: https://sourceforge.net/p/guliverkli2/code/HEAD/tree/src/subtitles/STS.cpp#l1447
        #   - tabulation and space : https://sourceforge.net/p/guliverkli2/code/HEAD/tree/src/subtitles/STS.cpp#l1172
        sub_styles: Dict[str, AssStyle] = {
            style.name.lstrip("\t ").lstrip("*"): AssStyle(
                style.fontname.lstrip("\t "), 700 if style.bold else 400, style.italic
            )
            for style in self.subtitle.styles
        }
        sub_wrap_style = WrapStyle(self.subtitle.wrap_style)

        for i, line in enumerate(self.subtitle.events):
            if isinstance(line, Dialogue):
                try:
                    original_line_style = sub_styles[line.style]
                except KeyError:
                    raise ValueError(
                        f'Error: Unknown style "{line.style}" on line {i+1}. You need to correct the .ass file.'
                    )

                tags = parse_line(line.text)

                # Copy the original_line_style
                line_style = AssStyle(
                    original_line_style.fontname,
                    original_line_style.weight,
                    original_line_style.italic,
                )
                current_style = AssStyle(
                    original_line_style.fontname,
                    original_line_style.weight,
                    original_line_style.italic,
                )

                self._set_used_styles(
                    used_styles,
                    tags,
                    i + 1,
                    sub_styles,
                    original_line_style,
                    line_style,
                    current_style,
                    sub_wrap_style,
                )

        return used_styles



import freetype
import logging
from io import BufferedReader
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union
from fontTools.ttLib.ttFont import TTFont
from fontTools.ttLib.tables._n_a_m_e import NameRecord
from fontTools.varLib.instancer import AxisLimits, axisValuesFromAxisLimits
from fontTools.varLib.instancer.names import (
    _isRibbi,
    _sortAxisValues,
    ELIDABLE_AXIS_VALUE_NAME,
    NameID,
)
from struct import error as struct_error

_logger = logging.getLogger(__name__)


class ParseFont:
    @staticmethod
    def get_var_font_family_prefix(font: TTFont) -> str:
        """
        From: https://github.com/fonttools/fonttools/blob/3c4cc71504774d1ae4f1e59e3ef3b97e194c1c91/Lib/fontTools/varLib/instancer/names.py#L267-L269

        Parameters:
            font (TTFont): An fontTools object
        Returns:
            The family name prefix.
        """
        family_prefix = ParseFont.get_name_by_id(
            NameID.TYPOGRAPHIC_FAMILY_NAME, font["name"].names
        ) or ParseFont.get_name_by_id(NameID.FAMILY_NAME, font["name"].names)

        return family_prefix

    @staticmethod
    def get_var_font_family_fullname(
        font: TTFont, axis_value_tables
    ) -> Tuple[str, str]:
        """
        From:
            - https://github.com/fonttools/fonttools/blob/3c4cc71504774d1ae4f1e59e3ef3b97e194c1c91/Lib/fontTools/varLib/instancer/names.py#L195-L232
            - https://github.com/fonttools/fonttools/blob/3c4cc71504774d1ae4f1e59e3ef3b97e194c1c91/Lib/fontTools/varLib/instancer/names.py#L267-L305

        Parameters:
            font (TTFont): An fontTools object
            axis_value_tables: List of axis value
        Returns:
            The family name and the fullname for an axis_value_tables.
        """
        if "STAT" not in font:
            raise ValueError(
                "Cannot get family and fullname since there is no STAT table."
            )

        stat = font["STAT"].table

        axis_value_tables = [
            v for v in axis_value_tables if not v.Flags & ELIDABLE_AXIS_VALUE_NAME
        ]

        axisValueNameIDs = [a.ValueNameID for a in axis_value_tables]
        ribbiNameIDs = [n for n in axisValueNameIDs if _isRibbi(font["name"], n)]
        nonRibbiNameIDs = [n for n in axisValueNameIDs if n not in ribbiNameIDs]
        elidedNameID = stat.ElidedFallbackNameID
        elidedNameIsRibbi = _isRibbi(font["name"], elidedNameID)

        subFamilyName = " ".join(
            ParseFont.get_name_by_id(n, font["name"].names) for n in ribbiNameIDs
        )
        if nonRibbiNameIDs:
            typoSubFamilyName = " ".join(
                ParseFont.get_name_by_id(n, font["name"].names)
                for n in axisValueNameIDs
            )
        else:
            typoSubFamilyName = None

        # If neither subFamilyName and typographic SubFamilyName exist,
        # we will use the STAT's elidedFallbackName
        if not typoSubFamilyName and not subFamilyName:
            if elidedNameIsRibbi:
                subFamilyName = ParseFont.get_name_by_id(
                    elidedNameID, font["name"].names
                )
            else:
                typoSubFamilyName = ParseFont.get_name_by_id(
                    elidedNameID, font["name"].names
                )

        familyNameSuffix = " ".join(
            ParseFont.get_name_by_id(n, font["name"].names) for n in nonRibbiNameIDs
        )

        currentFamilyName = ParseFont.get_var_font_family_prefix(font)

        nameIDs = {
            NameID.FAMILY_NAME: currentFamilyName,
            NameID.SUBFAMILY_NAME: subFamilyName or "Regular",
        }
        if typoSubFamilyName:
            nameIDs[
                NameID.FAMILY_NAME
            ] = f"{currentFamilyName} {familyNameSuffix}".strip()
            nameIDs[NameID.TYPOGRAPHIC_FAMILY_NAME] = currentFamilyName
            nameIDs[NameID.TYPOGRAPHIC_SUBFAMILY_NAME] = typoSubFamilyName

        newFamilyName = (
            nameIDs.get(NameID.TYPOGRAPHIC_FAMILY_NAME) or nameIDs[NameID.FAMILY_NAME]
        )
        newStyleName = (
            nameIDs.get(NameID.TYPOGRAPHIC_SUBFAMILY_NAME)
            or nameIDs[NameID.SUBFAMILY_NAME]
        )

        nameIDs[NameID.FULL_FONT_NAME] = f"{newFamilyName} {newStyleName}"

        return nameIDs[NameID.FAMILY_NAME], nameIDs[NameID.FULL_FONT_NAME]

    @staticmethod
    def get_axis_value_from_coordinates(font: TTFont, coordinates: Dict) -> List:
        """
        Parameters:
            font (TTFont): An fontTools object
            coordinates (Dict): An coordinates from an NamedInstance
        Returns:
            An list of each Axis Value for an coordinates.
        """

        if "STAT" not in font:
            raise ValueError("Cannot get axis_value since there is no STAT table.")
        stat = font["STAT"].table

        axisLimits = AxisLimits(coordinates).populateDefaults(font)
        axis_value_tables = axisValuesFromAxisLimits(stat, axisLimits)

        return _sortAxisValues(axis_value_tables)

    @staticmethod
    def sort_naming_table(names: List[NameRecord]) -> List[NameRecord]:
        """
        Parameters:
            names (List[NameRecord]): Naming table
        Returns:
            The sorted naming table
        """

        def isEnglish(name: NameRecord) -> bool:
            # From https://gitlab.freedesktop.org/fontconfig/fontconfig/-/blob/d863f6778915f7dd224c98c814247ec292904e30/src/fcfreetype.c#L1111-1125
            return (name.platformID, name.langID) in ((1, 0), (3, 0x409))

        # From	https://gitlab.freedesktop.org/fontconfig/fontconfig/-/blob/d863f6778915f7dd224c98c814247ec292904e30/src/fcfreetype.c#L1078
        # 		https://github.com/freetype/freetype/blob/b98dd169a1823485e35b3007ce707a6712dcd525/include/freetype/ttnameid.h#L86-L91
        PLATFORM_ID_APPLE_UNICODE = 0
        PLATFORM_ID_MACINTOSH = 1
        PLATFORM_ID_ISO = 2
        PLATFORM_ID_MICROSOFT = 3
        PLATFORM_ID_ORDER = [
            PLATFORM_ID_MICROSOFT,
            PLATFORM_ID_APPLE_UNICODE,
            PLATFORM_ID_MACINTOSH,
            PLATFORM_ID_ISO,
        ]

        return sorted(
            names,
            key=lambda name: (
                PLATFORM_ID_ORDER.index(name.platformID),
                name.nameID,
                name.platEncID,
                -isEnglish(name),
                name.langID,
            ),
        )

    @staticmethod
    def get_name_by_id(nameID: int, names: List[NameRecord]) -> str:
        """
        Parameters:
            nameID (int): ID of the name you search
            names (List[NameRecord]): Naming table
        Returns:
            The decoded name
        """

        names = list(filter(lambda name: name.nameID == nameID, names))
        names = ParseFont.sort_naming_table(names)

        for name in names:
            try:
                unistr = name.toUnicode()
            except UnicodeDecodeError:
                continue

            return unistr

    @staticmethod
    def get_font_postscript_property(
        font_path: str, font_index: int
    ) -> Union[str, None]:
        """
        Parameters:
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            is_italic, weight
        """
        try:
            # Like libass, we use freetype: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L326
            postscriptNameByte = freetype.Face(
                Path(font_path).open("rb"), font_index
            ).postscript_name
        except OSError:
            _logger.warning(
                f'Error: Please report this error on github. Attach this font "{font_path}" in your issue and say that the postscript has not been correctly decoded'
            )

        if postscriptNameByte is not None:
            try:
                postscriptName = postscriptNameByte.decode("ASCII")
            except UnicodeDecodeError:
                _logger.warning(
                    f'Error: Please report this error on github. Attach this font "{font_path}" in your issue and say that the postscript has not been correctly decoded'
                )

            if postscriptName is not None:
                return postscriptName

        return None

    @staticmethod
    def get_font_italic_bold_property(
        font: TTFont, font_path: str, font_index: int
    ) -> Tuple[bool, int]:
        """
        Parameters:
            font (TTFont): An fontTools object
            font_path (str): Font path.
            font_index (int): Font index.
        Returns:
            is_italic, weight
        """

        def get_font_italic_bold_property_with_freetype(
            font_path: str, font_index: int
        ) -> Tuple[bool, int]:
            font = freetype.Face(Path(font_path).open("rb"), font_index)
            # From: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L318
            is_italic = bool(
                font.style_flags & freetype.ft_enums.ft_style_flags.FT_STYLE_FLAG_ITALIC
            )
            # From: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_font.c#L523
            weight = (
                700
                if bool(
                    font.style_flags
                    & freetype.ft_enums.ft_style_flags.FT_STYLE_FLAG_BOLD
                )
                else 400
            )

            return is_italic, weight

        if "OS/2" in font:
            try:
                # https://docs.microsoft.com/en-us/typography/opentype/spec/os2#fss
                is_italic = bool(font["OS/2"].fsSelection & 1)

                # https://docs.microsoft.com/en-us/typography/opentype/spec/os2#usweightclass
                weight = font["OS/2"].usWeightClass
            except struct_error:
                is_italic, weight = get_font_italic_bold_property_with_freetype(
                    font_path, font_index
                )
        else:
            is_italic, weight = get_font_italic_bold_property_with_freetype(
                font_path, font_index
            )

        return is_italic, weight

    @staticmethod
    def get_font_family_fullname_property(
        names: List[NameRecord],
    ) -> Tuple[Set[str], Set[str]]:
        """
        Parameters:
            names (List[NameRecord]): Naming table
        Returns:
            All decoded families and fullnames that are from microsoft platform
        """

        # https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L258-L344
        MAX = 100
        families = fullnames = set()

        for name in names:

            if name.platformID == 3 and (name.nameID == 1 or name.nameID == 4):

                try:
                    # Since we use the windows platform, we can simply use utf_16_be to decode the string: https://docs.microsoft.com/en-us/typography/opentype/spec/name#platform-specific-encoding-and-language-ids-windows-platform-platform-id-3
                    # Even libass always use utf_16_be: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L283
                    nameStr = name.string.decode("utf_16_be")
                except UnicodeDecodeError:
                    continue

                if name.nameID == 1 and len(families) < MAX:
                    families.add(nameStr)
                elif name.nameID == 4 and len(fullnames) < MAX:
                    fullnames.add(nameStr)

        return families, fullnames

    @staticmethod
    def is_file_truetype_collection(file: BufferedReader) -> bool:
        file.seek(0)
        return b"ttcf" == file.read(4)

    @staticmethod
    def is_file_truetype(file: BufferedReader) -> bool:
        file.seek(0)
        return b"\x00\x01\x00\x00\x00" == file.read(5)

    @staticmethod
    def is_file_opentype(file: BufferedReader) -> bool:
        file.seek(0)
        return b"OTTO" == file.read(4)

    @staticmethod
    def is_file_font(filepath: str) -> bool:
        with open(filepath, "rb") as fontFile:
            return (
                ParseFont.is_file_truetype(fontFile)
                or ParseFont.is_file_opentype(fontFile)
                or ParseFont.is_file_truetype_collection(fontFile)
            )



from typing import Set


class UsageData:

    characters_used: Set[str]
    lines: Set[int]

    def __init__(
        self,
        characters_used: Set[str],
        lines: Set[int],
    ):
        self.characters_used = characters_used
        self.lines = lines

    @property
    def ordered_lines(self):
        lines = list(self.lines)
        lines.sort()
        return lines

    def __eq__(self, other: "UsageData"):
        return (self.characters_used, self.lines) == (
            other.characters_used,
            other.lines,
        )

    def __repr__(self):
        return f'characters_used: "{self.characters_used}", lines: "{self.lines}"'



import logging
import os
import shutil
from .ass_style import AssStyle
from .parse_font import ParseFont
from .font import Font
from .font_loader import FontLoader
from .font_result import FontResult
from ._version import __version__
from datetime import date
from fontTools.ttLib.ttFont import TTFont
from fontTools.ttLib.ttCollection import TTCollection
from fontTools.varLib import instancer
from pathlib import Path
from typing import List, Set, Tuple, Union

_logger = logging.getLogger(__name__)


class Helpers:
    @staticmethod
    def get_used_font_by_style(
        font_collection: Set[Font], style: AssStyle, search_by_family_name: bool = True
    ) -> Union[FontResult, None]:
        """
        Parameters:
            font_collection (Set[Font]): Font collection
            style (AssStyle): An AssStyle
            search_by_family_name (bool):
                If true, it will search the font by it's family name.
                If false, it will search the font by it's exact_name (fullname or postscript name).
        Returns:
            Ordered list of the font that match the best to the AssStyle
        """
        fonts_match: List[Tuple[int, Font]] = []

        for font in font_collection:
            if (search_by_family_name and style.fontname in font.family_names) or (
                not search_by_family_name and style.fontname in font.exact_names
            ):

                weight_compare = abs(style.weight - font.weight)

                if (style.weight - font.weight) > 150:
                    weight_compare -= 120

                # Thanks to rcombs@github: https://github.com/libass/libass/issues/613#issuecomment-1102994528
                weight_compare = (
                    ((((weight_compare << 3) + weight_compare) << 3)) + weight_compare
                ) >> 8

                fonts_match.append((weight_compare, font))

        # The last sort parameter (font.weight) is totally optional.
        # In VSFilter, when the weight_compare is the same between 2 fonts, it will take the first one, so the order is random.
        # Because of that, VSFilter can not always display the same font.
        # We also priorize Truetype/Opentype font over variable fonts (parameter is_var)
        fonts_match.sort(
            key=lambda font: (
                font[0],
                -(style.italic == font[1].italic),
                font[1].is_var,
                font[1].weight,
            )
        )

        if len(fonts_match) > 0:
            match = fonts_match[0][1]
            mismatch_italic = not (match.italic == style.italic)
            mismatch_bold = not (-150 < match.weight - style.weight < 150)

            font_result = FontResult(match, mismatch_bold, mismatch_italic)
            _logger.info(f"Found '{style.fontname}' at '{font_result.font.filename}'")
            return font_result
        elif search_by_family_name:
            return Helpers.get_used_font_by_style(
                font_collection, style, search_by_family_name=False
            )
        else:
            _logger.error(f"Could not find font '{style.fontname}'")
            return None

    @staticmethod
    def copy_font_to_directory(
        font_collection: Set[Font],
        output_directory: Path,
        create_directory_if_doesnt_exist: bool = True,
        convert_variable_font_into_truetype_collection: bool = True,
    ):
        """
        Parameters:
            font_collection (Set[Font]): Font collection
            output_directory (Path): The directory where the font are going to be save
            create_directory_if_doesnt_exist (bool):
                If true and the output_directory doesn't exist, it will be created
                If false and the output_directory doesn't exist, an exception will be raised
            convert_variable_font_into_truetype_collection (bool):
                If true, it will convert the variable font into an truetype collection font
                    It is usefull, because libass doesn't support variation font: https://github.com/libass/libass/issues/386
                    It convert it in a format that libass support
                If false, it won't do anything special. The variable font will be copied like any other font.
        """

        if not os.path.exists(output_directory) and create_directory_if_doesnt_exist:
            os.makedirs(output_directory)

        for font in font_collection:

            if font.is_var and convert_variable_font_into_truetype_collection:
                # We take the first result, but it doesn't matter
                font = Helpers.variable_font_to_collection(font.filename, os.getcwd())[
                    0
                ]

            # Don't overwrite fonts
            if not os.path.isfile(
                os.path.join(output_directory, Path(font.filename).name)
            ):
                shutil.copy(font.filename, output_directory)

    @staticmethod
    def variable_font_to_collection(
        fontpath: str, output_directory: str, cache_generated_font: bool = True
    ) -> List[Font]:
        """
        Parameters:
            fontpath (str): The path to the variable font that need to be converted
            output_directory (str): Path where to save the generated font
            cache_generated_font (bool):  Converting an variable font into an collection font is a slow process. Caching the result boost the performance.
                If true, then the generated font will be cached.
                If false, then the generated font won't be cached.
        Returns:
            List of Font that represent the truetype collection font generated
        """
        font_collection = TTCollection()
        ttFont = TTFont(fontpath)

        family_prefix = ParseFont.get_var_font_family_prefix(ttFont)

        for instance in ttFont["fvar"].instances:
            generated_font = instancer.instantiateVariableFont(
                ttFont, instance.coordinates
            )

            axis_value_tables = ParseFont.get_axis_value_from_coordinates(
                ttFont, instance.coordinates
            )
            family_name, full_font_name = ParseFont.get_var_font_family_fullname(
                ttFont, axis_value_tables
            )

            generated_font["name"].setName(family_name, 1, 3, 1, 0x409)
            generated_font["name"].setName(full_font_name, 2, 3, 1, 0x409)
            generated_font["name"].setName(
                f"FontCollector v {__version__}:{full_font_name}:{date.today()}",
                3,
                3,
                1,
                0x409,
            )
            generated_font["name"].setName(full_font_name, 4, 3, 1, 0x409)

            font_collection.fonts.append(generated_font)

        savepath = os.path.join(output_directory, f"{family_prefix}.ttc")
        font_collection.save(savepath)

        generated_font = Font.from_font_path(savepath)
        if cache_generated_font:
            for font in generated_font:
                FontLoader.add_generated_font(font)

        return generated_font



import os
import pickle
from .font import Font
from matplotlib import font_manager
from pathlib import Path
from tempfile import gettempdir
from typing import List, Set


class FontLoader:
    """
    system_fonts: Font installed on the system
    additional_fonts: Font added by the user
    generated_fonts: Contain all the generated font by Helpers.variable_font_to_collection.
    """

    system_fonts: Set[Font]
    additional_fonts: Set[Font]

    def __init__(
        self, additional_fonts_path: List[Path] = [], use_system_font: bool = True
    ):

        if use_system_font:
            self.system_fonts = FontLoader.load_system_fonts()
        else:
            self.system_fonts = set()

        self.additional_fonts = FontLoader.load_additional_fonts(additional_fonts_path)

    @property
    def fonts(self) -> Set[Font]:
        """
        Get all the fonts
        """
        return self.system_fonts.union(FontLoader.load_generated_fonts()).union(
            self.additional_fonts
        )

    def add_additional_font(self, font_path: Path):
        """
        Parameters:
            font_path (str): Font path. The font can be a .ttf, .otf or .ttc
                If you need to use woff font, you will need to decompress them.
                See fontTools documentation to know how to do it: https://fonttools.readthedocs.io/en/latest/ttLib/woff2.html#fontTools.ttLib.woff2.decompress
        """
        self.additional_fonts.update(FontLoader.load_additional_fonts([font_path]))

    @staticmethod
    def add_generated_font(font: Font):
        """
        Parameters:
            font (Font): Generated font by Helpers.variable_font_to_collection
        """

        generated_fonts = FontLoader.load_generated_fonts()
        generated_fonts.add(font)
        FontLoader.save_generated_fonts(generated_fonts)

    @staticmethod
    def load_system_fonts() -> Set[Font]:
        system_fonts: Set[Font] = set()
        fonts_paths: Set[str] = set(font_manager.findSystemFonts())
        system_font_cache_file = FontLoader.get_system_font_cache_file_path()

        if os.path.exists(system_font_cache_file):

            with open(system_font_cache_file, "rb") as file:
                cached_fonts: Set[Font] = pickle.load(file)

            cached_paths = set(map(lambda font: font.filename, cached_fonts))

            # Remove font that aren't anymore installed
            removed = cached_paths.difference(fonts_paths)
            system_fonts = set(
                filter(lambda font: font.filename not in removed, cached_fonts)
            )

            # Add font that have been installed since last execution
            added = fonts_paths.difference(cached_paths)
            for font_path in added:
                system_fonts.update(Font.from_font_path(font_path))

            # If there is a change, update the cache file
            if len(added) > 0 or len(removed) > 0:
                with open(system_font_cache_file, "wb") as file:
                    pickle.dump(system_fonts, file)

        else:
            # Since there is no cache file, load the font
            for font_path in fonts_paths:
                system_fonts.update(Font.from_font_path(font_path))

            # Save the font into the cache file
            with open(system_font_cache_file, "wb") as file:
                pickle.dump(system_fonts, file)

        return system_fonts

    @staticmethod
    def load_generated_fonts() -> Set[Font]:
        generated_fonts: Set[Font] = set()
        generated_font_cache_file = FontLoader.get_generated_font_cache_file_path()

        if os.path.exists(generated_font_cache_file):
            with open(generated_font_cache_file, "rb") as file:
                cached_fonts: Set[Font] = pickle.load(file)

            generated_fonts = set(
                filter(lambda font: os.path.exists(font.filename), cached_fonts)
            )

        return generated_fonts

    @staticmethod
    def load_additional_fonts(additional_fonts_path: List[Path]) -> Set[Font]:
        additional_fonts: Set[Font] = set()

        for font_path in additional_fonts_path:
            if os.path.isfile(font_path):
                additional_fonts.update(Font.from_font_path(font_path))
            elif os.path.isdir(font_path):
                for path in font_manager.findSystemFonts(fontpaths=str(font_path)):
                    additional_fonts.update(Font.from_font_path(path))
            else:
                raise FileNotFoundError(f"The file {font_path} is not reachable")
        return additional_fonts

    @staticmethod
    def save_generated_fonts(generated_fonts: Set[Font]):
        generated_font_cache_file = FontLoader.get_generated_font_cache_file_path()
        with open(generated_font_cache_file, "wb") as file:
            pickle.dump(generated_fonts, file)

    @staticmethod
    def discard_system_font_cache():
        system_font_cache = FontLoader.get_system_font_cache_file_path()
        if os.path.isfile(system_font_cache):
            os.remove(system_font_cache)

    @staticmethod
    def discard_generated_font_cache():
        generated_font_cache = FontLoader.get_generated_font_cache_file_path()
        if os.path.isfile(generated_font_cache):
            os.remove(generated_font_cache)

    @staticmethod
    def get_system_font_cache_file_path() -> Path:
        tempDir = gettempdir()
        return Path(os.path.join(tempDir, "FontCollector_SystemFont.bin"))

    @staticmethod
    def get_generated_font_cache_file_path() -> Path:
        tempDir = gettempdir()
        return Path(os.path.join(tempDir, "FontCollector_GeneratedFont.bin"))



import logging
import shutil
import subprocess
from .font import Font
from os import path
from pathlib import Path
from typing import Sequence

_logger = logging.getLogger(__name__)


class Mkvpropedit:
    path: str = shutil.which("mkvpropedit")

    @staticmethod
    def is_mkv(filename: Path) -> bool:
        """
        Parameters:
            filename (Path): The file name. Example: "example.mkv"
        Returns:
            True if the mkv is an mkv file, false in any others cases
        Thanks to https://github.com/TypesettingTools/Myaamori-Aegisub-Scripts/blob/f2a52ee38eeb60934175722fa9d7f2c2aae015c6/scripts/fontvalidator/fontvalidator.py#L414
        """

        if not path.exists(filename):
            raise FileNotFoundError(f'The file "{filename}" does not exist.')

        with open(filename, "rb") as f:
            # From https://en.wikipedia.org/wiki/List_of_file_signatures
            return f.read(4) == b"\x1a\x45\xdf\xa3"

    @staticmethod
    def is_mkvpropedit_path_valid() -> bool:

        mkvpropeditOutput = subprocess.run(
            f"{Mkvpropedit.path} --version", capture_output=True, text=True
        )
        return mkvpropeditOutput.stdout.startswith("mkvpropedit")

    @staticmethod
    def delete_fonts_of_mkv(mkv_filename: Path) -> None:
        """
        Delete all mkv attached font
        Parameters:
            mkvFileName (Path): Path to mkvFile
        """

        if not Mkvpropedit.is_mkvpropedit_path_valid():
            raise FileNotFoundError(
                f'"{Mkvpropedit.path}" is not an valid path for Mkvpropedit. You need to correct your environnements variable or change the value of Mkvpropedit.path'
            )

        if not Mkvpropedit.is_mkv(mkv_filename):
            raise FileExistsError(f'The file "{mkv_filename}" is not an mkv file.')

        # We only want to remove ttf, ttc or otf file
        # This is from mpv: https://github.com/mpv-player/mpv/blob/305332f8a06e174c5c45c9c4547293502ac7ecdb/sub/sd_ass.c#L101

        mkvpropedit_args = [
            f'"{Mkvpropedit.path}"',
            f'"{mkv_filename}"',
            "--delete-attachment mime-type:application/x-truetype-font",
            "--delete-attachment mime-type:application/vnd.ms-opentype",
            "--delete-attachment mime-type:application/x-font-ttf",
            "--delete-attachment mime-type:application/x-font",
            "--delete-attachment mime-type:application/font-sfnt",
            "--delete-attachment mime-type:font/collection",
            "--delete-attachment mime-type:font/otf",
            "--delete-attachment mime-type:font/sfnt",
            "--delete-attachment mime-type:font/ttf",
        ]

        output = subprocess.run(
            " ".join(mkvpropedit_args), capture_output=True, text=True
        )

        if len(output.stderr) != 0:
            raise OSError(
                f"mkvpropedit reported an error when deleting the font in the mkv: {output.stderr}"
            )
        else:
            _logger.info(f'Successfully deleted fonts in mkv "{mkv_filename}')

    @staticmethod
    def merge_fonts_into_mkv(font_collection: Sequence[Font], mkv_filename: Path):
        """
        Parameters:
            font_collection (Sequence[Font]): All font needed to be merge in the mkv
            mkv_filename (Path): Mkv file path
        """
        if not Mkvpropedit.is_mkvpropedit_path_valid():
            raise FileNotFoundError(
                f'"{Mkvpropedit.path}" is not an valid path for Mkvpropedit. You need to correct your environnements variable or change the value of Mkvpropedit.path'
            )

        if not Mkvpropedit.is_mkv(mkv_filename):
            raise FileExistsError(f'The file "{mkv_filename}" is not an mkv file.')

        mkvpropedit_args = [
            f'"{Mkvpropedit.path}"',
            f'"{mkv_filename}"',
        ]

        for font in font_collection:
            mkvpropedit_args.append(f'--add-attachment "{font.filename}"')

        output = subprocess.run(
            " ".join(mkvpropedit_args), capture_output=True, text=True
        )

        if len(output.stderr) != 0:
            raise OSError(
                f"mkvpropedit reported an error when merging font into an mkv: {output.stderr}"
            )
        else:
            _logger.info(f'Successfully merging fonts into mkv "{mkv_filename}')



class AssStyle:
    """
    AssStyle is an instance that does not only represent "[V4+ Styles]" section of an .ass script.
    It also consider the tags \r, \i, \b and \fn
    """

    __fontname: str
    weight: int  # a.k.a bold
    italic: bool

    def __init__(
        self,
        fontname: str,
        weight: int,
        italic: bool,
    ):
        self.fontname = fontname
        self.weight = weight
        self.italic = italic

    @property
    def fontname(self):
        return self.__fontname

    @fontname.setter
    def fontname(self, value: str):
        self.__fontname = AssStyle.strip_fontname(value.lower())

    def __eq__(self, other: "AssStyle"):
        return (self.fontname, self.weight, self.italic) == (
            other.fontname,
            other.weight,
            other.italic,
        )

    def __hash__(self):
        return hash((self.fontname, self.weight, self.italic))

    def __repr__(self):
        return f'Fontname: "{self.fontname}", weight: "{self.weight}", italic: "{self.italic}"'

    @staticmethod
    def strip_fontname(fontName: str) -> str:
        """
        Inpired by:
            - https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_parse.c#L101
            - https://github.com/TypesettingTools/Myaamori-Aegisub-Scripts/blob/f2a52ee38eeb60934175722fa9d7f2c2aae015c6/scripts/fontvalidator/fontvalidator.py#L33-L37

        Parameters:
            fontName (str): The font name.
        Returns:
            The font without an @ at the beginning
        """
        if fontName.startswith("@"):
            return fontName[1:]
        else:
            return fontName



import logging
import sys
from .ass_document import AssDocument
from .font_loader import FontLoader
from .font_result import FontResult
from .helpers import Helpers
from .mkvpropedit import Mkvpropedit
from .parse_arguments import parse_arguments
from typing import List


_logger = logging.getLogger(__name__)


def main():
    (
        ass_files_path,
        output_directory,
        mkv_path,
        delete_fonts,
        additional_fonts,
        use_system_font,
    ) = parse_arguments()
    font_results: List[FontResult] = []
    font_collection = FontLoader(additional_fonts, use_system_font).fonts

    for ass_path in ass_files_path:
        subtitle = AssDocument.from_file(ass_path)
        _logger.info(f"Loaded successfully {ass_path}")
        styles = subtitle.get_used_style()

        nbr_font_not_found = 0

        for style, usage_data in styles.items():

            font_result = Helpers.get_used_font_by_style(font_collection, style)

            # Did not found the font
            if font_result is None:
                nbr_font_not_found += 1
                _logger.error(
                    f"Used on lines: {' '.join(str(line) for line in usage_data.ordered_lines)}"
                )
            else:
                font_results.append(font_result)

                if font_result.mismatch_bold:
                    _logger.warning(f"'{style.fontname}' does not have a bold variant.")
                if font_result.mismatch_italic:
                    _logger.warning(
                        f"'{style.fontname}' does not have an italic variant."
                    )

                if font_result.mismatch_bold or font_result.mismatch_italic:
                    _logger.warning(
                        f"Used on lines: {' '.join(str(line) for line in usage_data.ordered_lines)}"
                    )

                if (
                    len(
                        missing_glyphs := font_result.font.get_missing_glyphs(
                            usage_data.characters_used
                        )
                    )
                    > 0
                ):
                    _logger.warning(
                        f"'{style.fontname}' is missing the following glyphs used: {missing_glyphs}"
                    )

        if nbr_font_not_found == 0:
            _logger.info(f"All fonts found")
        else:
            _logger.info(f"{nbr_font_not_found} fonts could not be found.")

    # Remove duplicate font
    font_found = list({font.font.filename: font.font for font in font_results}.values())

    if mkv_path is not None:
        if delete_fonts:
            Mkvpropedit.delete_fonts_of_mkv(mkv_path)
        Mkvpropedit.merge_fonts_into_mkv(font_found, mkv_path)
    else:
        Helpers.copy_font_to_directory(font_found, output_directory)


if __name__ == "__main__":
    sys.exit(main())



import os
from .font import Font
from .font_loader import FontLoader
from .mkvpropedit import Mkvpropedit
from argparse import ArgumentParser
from pathlib import Path
from typing import List, Set, Tuple, Union


def _parse_input_file(ass_input: List[Path]) -> List[Path]:

    if len(ass_input) == 0:
        return [Path(file) for file in os.listdir(os.getcwd()) if file.endswith(".ass")]

    ass_files_path = []
    for input in ass_input:
        if os.path.isfile(input):
            if input.name.endswith(".ass"):
                ass_files_path.append(Path(input))
            else:
                raise FileExistsError("Error: The input file is not an .ass file.")
        elif os.path.isdir(input):
            for file in os.listdir(input):
                if file.endswith(".ass"):
                    ass_files_path.append(Path(os.path.join(input, file)))
        else:
            raise FileNotFoundError(
                f"Error: The input file/folder {input.name} does not exist."
            )

    return ass_files_path


def parse_arguments() -> Tuple[
    List[Path],
    Path,
    Union[Path, None],
    bool,
    Set[Font],
    bool,
]:
    """
    Returns:
        ass_files_path, output_directory, mkv_path, delete_fonts, additional_fonts, use_system_fonts
    """
    parser = ArgumentParser(
        description="FontCollector for Advanced SubStation Alpha file."
    )
    parser.add_argument(
        "--input",
        "-i",
        nargs="*",
        type=Path,
        required=True,
        help="""
    Subtitles file. Must be an ASS file/directory. You can specify more than one .ass file/path. If no argument is specified, it will take all the font in the current path.
    """,
    )
    parser.add_argument(
        "-mkv",
        type=Path,
        help="""
    Video where the fonts will be merge. Must be a Matroska file.
    """,
    )
    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        default=os.getcwd(),
        help="""
    Destination path of the font. If -o and -mkv aren't specified, it will be the current path.
    """,
    )
    parser.add_argument(
        "-mkvpropedit",
        type=Path,
        help="""
    Path to mkvpropedit.exe if not in variable environments. If -mkv is not specified, it will do nothing.
    """,
    )
    parser.add_argument(
        "--delete-fonts",
        "-d",
        action="store_true",
        help="""
    If -d is specified, it will delete the font attached to the mkv before merging the new needed font. If -mkv is not specified, it will do nothing.
    """,
    )
    parser.add_argument(
        "--additional-fonts",
        nargs="+",
        type=Path,
        help="""
    May be a directory containing font files or a single font file. You can specify more than one additional-fonts.
    """,
    )
    parser.add_argument(
        "--exclude-system-fonts",
        action="store_false",
        help="""
    If specified, FontCollector won't use the system font to find the font used by an .ass file.
    """,
    )

    args = parser.parse_args()

    # Parse args
    ass_files_path = _parse_input_file(args.input)

    output_directory = args.output

    mkv_path = args.mkv
    if mkv_path:
        if args.mkvpropedit is not None:
            Mkvpropedit.path = args.mkvpropedit

    delete_fonts = args.delete_fonts

    if args.additional_fonts is not None:
        additional_fonts = FontLoader.load_additional_fonts(args.additional_fonts)
    else:
        additional_fonts = set()

    use_system_fonts = args.exclude_system_fonts

    return (
        ass_files_path,
        output_directory,
        mkv_path,
        delete_fonts,
        additional_fonts,
        use_system_fonts,
    )



import os
import re
import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install


here = os.path.abspath(os.path.dirname(__file__))


class PostDevelopCommand(develop):
    """
    Post-installation for development mode.
    From https://stackoverflow.com/a/36902139/15835974
    """

    def run(self):
        develop.run(self)
        from font_collector import FontLoader

        FontLoader.discard_generated_font_cache()
        FontLoader.discard_system_font_cache()


class PostInstallCommand(install):
    """
    Post-installation for installation mode.
    From https://stackoverflow.com/a/36902139/15835974
    """

    def run(self):
        install.run(self)
        from font_collector import FontLoader

        FontLoader.discard_generated_font_cache()
        FontLoader.discard_system_font_cache()


def read(*parts):
    with open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = [\'\"](.+)[\'\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="FontCollector",
    url="https://github.com/moi15moi/FontCollector/",
    project_urls={
        "Source": "https://github.com/moi15moi/FontCollector/",
        "Tracker": "https://github.com/moi15moi/FontCollector/issues/",
    },
    author="moi15moi",
    author_email="moi15moismokerlolilol@gmail.com",
    description="FontCollector for Advanced SubStation Alpha file.",
    long_description_content_type="text/markdown",
    version=find_version("font_collector", "_version.py"),
    packages=["font_collector"],
    python_requires=">=3.8",
    install_requires=[
        "ass",
        "ass-tag-analyzer",
        "fontTools>=4.38.0",
        "freetype-py",
        "matplotlib>=3.6",
    ],
    entry_points={"console_scripts": ["fontcollector=font_collector.__main__:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Other Audience",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    license="GNU LGPL 3.0 or later",
    cmdclass={
        "develop": PostDevelopCommand,
        "install": PostInstallCommand,
    },
)



from .font import Font


class FontResult:

    font: Font
    mismatch_bold: bool
    mismatch_italic: bool

    def __init__(
        self,
        font: Font,
        mismatch_bold: bool,
        mismatch_italic: bool,
    ):
        self.font = font
        self.mismatch_bold = mismatch_bold
        self.mismatch_italic = mismatch_italic

    def __repr__(self):
        return f'font: "{self.font}", mismatch_bold: "{self.mismatch_bold}", mismatch_italic: "{self.mismatch_italic}"'



import logging
from .ass_document import AssDocument
from .ass_style import AssStyle
from .font_loader import FontLoader
from .font_result import FontResult
from .font import Font
from .helpers import Helpers
from .mkvpropedit import Mkvpropedit
from .usage_data import UsageData
from ._version import __version__

# Set dependency log level to ERROR
from matplotlib import set_loglevel

set_loglevel("ERROR")

from fontTools.misc.loggingTools import configLogger

configLogger(level="ERROR")

# Set our default logger
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)

_formatter = logging.Formatter("%(levelname)s - %(message)s")

_handler = logging.StreamHandler()
_handler.setLevel(logging.INFO)
_handler.setFormatter(_formatter)

_logger.addHandler(_handler)


def set_loglevel(level: int):
    """
    Parameters:
        level (int): An level from logging module (For more detail, see: https://docs.python.org/3/library/logging.html#logging-levels)
    """
    _logger.setLevel(level)
    _handler.setLevel(level)



import logging
from .parse_font import ParseFont
from fontTools.ttLib.ttFont import TTFont
from fontTools.ttLib.ttCollection import TTCollection
from typing import List, Sequence, Set

_logger = logging.getLogger(__name__)


class Font:
    filename: str
    font_index: int
    __family_names: Set[str]
    weight: int
    italic: bool
    __exact_names: Set[
        str
    ]  # if the font is a TrueType, it will be the "full_name". if the font is a OpenType, it will be the "postscript name"
    is_var: bool

    def __init__(
        self,
        filename: str,
        font_index: int,
        family_names: Set[str],
        weight: int,
        italic: bool,
        exact_names: Set[str],
        is_var: bool,
    ):
        self.filename = filename
        self.font_index = font_index
        self.family_names = family_names
        self.weight = weight
        self.italic = italic
        self.exact_names = exact_names
        self.is_var = is_var

    @classmethod
    def from_font_path(cls, font_path: str) -> List["Font"]:
        """
        Parameters:
            font_path (str): Font path. The font can be a .ttf, .otf or .ttc
        Returns:
            An Font object that represent the file at the font_path
        """
        ttFonts: List[TTFont] = []
        fonts: List[Font] = []

        with open(font_path, "rb") as font_file:
            if ParseFont.is_file_truetype_collection(font_file):
                ttFonts.extend(TTCollection(font_path).fonts)
            elif ParseFont.is_file_truetype(font_file) or ParseFont.is_file_opentype(
                font_file
            ):
                ttFonts.append(TTFont(font_path))
            else:
                raise FileExistsError(
                    f'The file "{font_path}" is not a valid font file'
                )

        # Read font attributes
        for font_index, ttFont in enumerate(ttFonts):

            families = set()
            exact_names = set()

            # If is Variable Font, else "normal" font
            if is_var := "fvar" in ttFont:

                for instance in ttFont["fvar"].instances:
                    axis_value_tables = ParseFont.get_axis_value_from_coordinates(
                        ttFont, instance.coordinates
                    )
                    (
                        family_name,
                        full_font_name,
                    ) = ParseFont.get_var_font_family_fullname(
                        ttFont, axis_value_tables
                    )

                    families.add(family_name)
                    families.add(full_font_name)

            else:
                # From https://github.com/fonttools/fonttools/discussions/2619
                isTrueType = False
                if "glyf" in ttFont:
                    isTrueType = True

                families, fullnames = ParseFont.get_font_family_fullname_property(
                    ttFont["name"].names
                )

                # This is something like: https://github.com/libass/libass/blob/a2b39cde4ecb74d5e6fccab4a5f7d8ad52b2b1a4/libass/ass_fontselect.c#L303-L311
                if len(families) == 0:
                    familyName = ParseFont.get_name_by_id(1, ttFont["name"].names)

                    if familyName:
                        families.add(familyName)
                    else:
                        # Skip the font since it is invalid
                        _logger.info(
                            f'Warning: The index {font_index} of the font "{font_path}" does not contain an valid family name. The font index will be ignored.'
                        )
                        continue

                if isTrueType:
                    exact_names = fullnames
                else:
                    # If not TrueType, it is OpenType

                    postscript_name = ParseFont.get_font_postscript_property(
                        font_path, font_index
                    )
                    if postscript_name is not None:
                        exact_names.add(postscript_name)

            is_italic, weight = ParseFont.get_font_italic_bold_property(
                ttFont, font_path, font_index
            )
            fonts.append(
                Font(
                    font_path,
                    font_index,
                    families,
                    weight,
                    is_italic,
                    exact_names,
                    is_var,
                )
            )

        return fonts

    @property
    def family_names(self):
        return self.__family_names

    @family_names.setter
    def family_names(self, family_names: List[str]):
        self.__family_names = set([family_name.lower() for family_name in family_names])

    @property
    def exact_names(self):
        return self.__exact_names

    @exact_names.setter
    def exact_names(self, exact_names: List[str]):
        self.__exact_names = set([exact_name.lower() for exact_name in exact_names])

    def __eq__(self, other: "Font"):
        return (self.family_names, self.weight, self.italic, self.exact_names) == (
            other.family_names,
            other.weight,
            other.italic,
            other.exact_names,
        )

    def __hash__(self):
        return hash(
            (
                tuple(self.family_names),
                self.italic,
                self.weight,
                tuple(self.exact_names),
            )
        )

    def __repr__(self):
        return f'Filename: "{self.filename}" Family_names: "{self.family_names}", Weight: "{self.weight}", Italic: "{self.italic}, Exact_names: "{self.exact_names}"'

    def get_missing_glyphs(self, text: Sequence[str]):

        ttFont = TTFont(self.filename, fontNumber=self.font_index)

        char_not_found: Set[str] = set()

        for char in text:
            char_found = False
            for table in ttFont["cmap"].tables:
                if ord(char) in table.cmap:
                    char_found = True
                    break
            if not char_found:
                char_not_found.add(char)

        return char_not_found
